/**
   This program evaluates (interprets) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

       Prog ::= Exp
              | '(' 'prog' (Fun | Exp)* Exp ')'

        Fun ::= '(' 'fun' VARIABLE Lambda ')'   // a function declaration

     Lambda ::= '(' 'lambda' VARIABLE* Exp ')'  // formal parameters followed by function body

        Exp ::= Apply
              | If
              | While
              | Var
              | Set
              | Begin
              | Print
              | BExp
              | AExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Apply ::= '(' 'apply' Exp Exp* ')'  // function value followed by actual parameters

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

        Set ::= '(' 'set' VARIABLE Exp ')'

      Begin ::= '(' 'begin' Exp* Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [-][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class Evaluate
{
   public static int DEBUG = 1;

   public static Environment globalEnv;  // reference to the global environment object

   public static Value eval(Tree tree) throws EvalException
   {
      return evaluateProg( tree );
   }//eval()


   // Evaluate a prog.
   private static Value evaluateProg(Tree tree) throws EvalException
   {
      Value result = null;

      // Instantiate the global environment
      // (it will always be at the end of the
      // environment chain).
      globalEnv = new Environment();

      // Check whick kind of Prog we have.
      if ( ! tree.getElement().equals("prog") )
      {
         // Evaluate the single expression.
         result = evaluateExp(tree, globalEnv);
      }
      else
      {
         // Evaluate each Fun or Exp in the Prog.
         // Any Fun will have the side effect of putting
         // a function name in the global environment.
         // Any Var expressions will have the side effect
         // of putting a variable in the environment chain.
         // Any Set expressions will have the side effect
         // of changing a value in the environment chain.
         // Any Print expressions will have the side effect
         // of printing an output.
         // Any other expressions would be pointless!
         for (int i = 0; i < tree.degree()-1; i++)
         {
            if ( tree.getSubTree(i).getElement().equals("fun") )
            {
               handleFun(tree.getSubTree(i), globalEnv);
            }
            else
            {
               evaluateExp(tree.getSubTree(i), globalEnv);
            }
         }

         // Evaluate the last expression and use its
         // value as the value of the prog expression.
         result = evaluateExp(tree.getSubTree(tree.degree()-1), globalEnv);
      }

      return result;
   }//evaluateProg()


   /**
      Handle a function definition. Notice that the return type
      is void since Fun isn't an expression.

      This method mutates the global environment object.
      The Value object put into the environment by this method
      has the tag "lambda" and its value field is a reference
      to the function's "lambda expression".
   */
   public static void handleFun(Tree tree, Environment env) throws EvalException
   {
      // get the function name
      String name = tree.getSubTree(0).getElement();
      // check if this function name has already been defined
      if ( env.definedLocal(name) )
      {
         throw new EvalException("function already exists: " + name);
      }
      // check if the definition really is a function
      if ( ! tree.getSubTree(1).getElement().equals("lambda"))
      {
         throw new EvalException("bad function definition: " + tree);
      }

      // get the "lambda expression" and create a function Value
      Value lambda = new Value( tree.getSubTree(1) );

      // add the <name, value> pair to the environment
      env.add(name, lambda);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes
   }//handleFun()


   // Evaluate an expression.
   public static Value evaluateExp(Tree tree, Environment env) throws EvalException
   {
      Value result = null;

      String node = tree.getElement();

      if ( node.equals("apply") )
         result = evaluateApply(tree, env);
      else if ( node.equals("if") )
         result = evaluateIf(tree, env);
      else if ( node.equals("while") )
         result = evaluateWhile(tree, env);
      else if ( node.equals("var") )
         result = evaluateVar(tree, env);
      else if ( node.equals("set") )
         result = evaluateSet(tree, env);
      else if ( node.equals("begin") )
         result = evaluateBegin(tree, env);
      else if ( node.equals("print") )
         result = evaluatePrint(tree, env);
      else if ( node.equals("&&") || node.equals("||")
             || node.equals("!") )
         result = evaluateBexp(tree, env);  // boolean expression
      else if ( node.equals("<")  || node.equals(">")
             || node.equals("<=") || node.equals(">=")
             || node.equals("==") || node.equals("!=") )
         result = evaluateRexp(tree, env);  // relational operator
      else if ( node.equals("+") || node.equals("-")
             || node.equals("*") || node.equals("/")
             || node.equals("%") || node.equals("^") )
         result = evaluateAexp(tree, env);  // arithmetic expression
      else if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
            result = new Value( node.equals("true") );
         else if ( node.matches("^[-]*[0-9][0-9]*") )
            result = new Value( Integer.parseInt( node ) );
         else if ( env.defined(node) )  // a variable
            result = env.lookUp( node );
         else  // runtime check
            throw new EvalException("undefined variable: " + node + "\n");
      }
      else
         throw new EvalException("invalid expression: " + tree + "\n");

      return result;
   }//evaluateExp()


   /**
      This method "applies" a function value to actual parameters.
      This method evaluates the body of the function in an environment
      that binds the actual parameter values (from this function application)
      to the formal parameters (from the function's lambda expression).
   */
   private static Value evaluateApply(Tree tree, Environment env) throws EvalException
   {
      // Evaluate the apply's first parameter to a function value,
/*1*/ Value funValue = evaluateExp(tree.getSubTree(0), env);
      // check that what we are applying really is a function
      if ( ! funValue.tag.equals(Value.LAMBDA_TAG))  // runtime check
      {
         throw new EvalException("bad function value: " + tree);
      }
      // get a reference to the function's "lambda expression".
      Tree lambda = funValue.valueL;

      // Check that the number of actual parameters
      // is equal to the number of formal parameters.
      // (Actually, all we really need to know is that
      // the number of actual parameters is at least
      // the number of formal parameters.)
      if ( tree.degree() != lambda.degree() )  // runtime check
      {
         throw new EvalException("wrong number of parameters: " + tree);
      }

      // Create a new environment object that is "nested"
      // in the global environment ("lexical scope") (or, is
      // "nested" in the current environment ("dynamic scope")).
      // This new environment will hold the bindings of each
      // actual parameter value to a formal paramter name.
      // (Note: Change globalEnv to env for dynamic scope.)
/*2*/ Environment localEnv = new Environment(globalEnv, "Function Activation");

      // Bind, in the new environment object, each actual parameter
      // value to a formal parameter name.
/*3*/ for (int i = 1; i < tree.degree(); i++) // iterate through the actual parameters
      {
         // Evaluate, using the current environment chain
         // (NOT the new local environment object),
         // an actual parameter expression to a value.
/*4*/    Value actualParamValue = evaluateExp(tree.getSubTree(i), env);

         // Retrieve, from within the lambda expression,
         // a formal parameter name.
/*5*/    String formalParamName = lambda.getSubTree(i-1).getElement();

         // Bind, in the new local environment object, the actual
         // paramter value to the formal parameter name.
/*6*/    localEnv.add(formalParamName, actualParamValue);
      }

      if (DEBUG > 0) System.out.println( localEnv + "\n" ); // for debugging purposes

      // Evaluate the body of the lambda expression using the
      // new environment (which contains the binding of the actual
      // parameter values to the function's formal parameter names).
/*7*/ Value result = evaluateExp(lambda.getSubTree(tree.degree()-1), localEnv);

      return result;
   }//evaluateApply()


   // Evaluate an if-expression.
   private static Value evaluateIf(Tree tree, Environment env) throws EvalException
   {
      if ( 3 != tree.degree() )  // runtime check
      {
         throw new EvalException("incorrect conditional expression: " + tree + "\n");
      }

      Value result = null;

      Value conditionalExp = evaluateExp(tree.getSubTree(0), env);
      // do a runtime type check
      if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
      {
         throw new EvalException("illegal boolean expression: " + tree);
      }

      if ( conditionalExp.valueB )
      {
         result = evaluateExp(tree.getSubTree(1), env);
      }
      else
      {
         result = evaluateExp(tree.getSubTree(2), env);
      }

      return result;
   }//evaluateIf()


   // Evaluate a while-loop expression.
   private static Value evaluateWhile(Tree tree, Environment env) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("incorrect while expression: " + tree + "\n");
      }

      Value result = null;

      // evaluate the boolean condition
      Value conditionalExp = evaluateExp(tree.getSubTree(0), env);
      // do a runtime type check
      if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
      {
         throw new EvalException("illegal boolean expression: " + tree);
      }

      while ( conditionalExp.valueB )
      {
         // evaluate the body of the loop (for its side effects)
         evaluateExp(tree.getSubTree(1), env);
         // re-evaluate the boolean condition
         conditionalExp = evaluateExp(tree.getSubTree(0), env);
         // do a runtime type check
         if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
         {
            throw new EvalException("illegal boolean expression: " + tree);
         }
      }

      // always return false for a while-loop expression
      result =  new Value( false );

      return result;
   }//evaluateWhile()


   // Evaluate a var expression.
   private static Value evaluateVar(Tree tree, Environment env) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = null;

      // get the variable
      String variable = tree.getSubTree(0).getElement();

      // check if this variable has already been declared
      // in the local environment
      if ( env.definedLocal(variable) )  // runtime check
      {
         throw new EvalException("varaible already declared: " + variable + "\n");
      }

      // get, and then evaluate, the expression
      Tree expr = tree.getSubTree(1);
      result = evaluateExp(expr, env);

      // declare the new, local, variable
      env.add(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//evaluateVar()


   // Evaluate a set expression.
   private static Value evaluateSet(Tree tree, Environment env) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = null;

      // get the variable
      String variable = tree.getSubTree(0).getElement();
      // check if this variable has already been declared
      if ( ! env.defined(variable) )  // runtime check
      {
         throw new EvalException("undefined variable: " + variable);
      }

      // get, and then evaluate, the expression
      Tree expr = tree.getSubTree(1);
      result = evaluateExp(expr, env);
      // update this variable in the environment
      env.update(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//evaluateSet()


   // Evaluate a begin expression.
   private static Value evaluateBegin(Tree tree, Environment env) throws EvalException
   {
      Value result = null;

      // Create a new Environment object chained to (or "nested in")
      // the previous ("outer") environment object.
      Environment newEnv = new Environment(env, "Local (begin)");

      // Evaluate each sub expression in the begin
      // expression (using the new environment chain).
      // The return value of each expression is
      // discarded, so any expression without a
      // side-effect is worthless.
      for (int i = 0; i < tree.degree()-1; i++)
      {
         evaluateExp(tree.getSubTree(i), newEnv);
      }

      // Evaluate the last expression and use its
      // value as the value of the begin expression.
      result = evaluateExp(tree.getSubTree(tree.degree()-1), newEnv);

      return result;
   }//evaluateBegin()


   // Evaluate a print expression.
   private static Value evaluatePrint(Tree tree, Environment env) throws EvalException
   {
      if ( 1 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = evaluateExp(tree.getSubTree(0), env);

      System.out.print( result + "\n" );

      return result;
   }//evaluatePrint()


   // Evaluate a boolean expression.
   private static Value evaluateBexp(Tree tree, Environment env) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value value = evaluateExp(tree.getSubTree(0), env);
      if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
      {
         throw new EvalException("not a boolean expression: "
                                  + tree.getSubTree(0) + "\n");
      }
      result = value.valueB;

      if ( node.equals("&&") )
      {
         if ( 2 > tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            if (result)
            {
               value = evaluateExp(tree.getSubTree(i), env);
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalException("not a boolean expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result = result && value.valueB;
            }
            else  // short circuit the evaluation of '&&'
            {
               result = false;
               break;
            }
         }
      }
      else if ( node.equals("||") )
      {
         if ( 2 > tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            if (! result)
            {
               value = evaluateExp(tree.getSubTree(i), env);
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalException("not a boolean expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result = result || value.valueB;
            }
            else  // short circuit the evaluation of '||'
            {
               result = true;
               break;
            }
         }
      }
      else if ( node.equals("!") )
      {
         if ( 1 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = ! result;
      }

      return new Value( result );
   }//evaluateBexp()


   // Evaluate a relational expression (which is a kind of boolean expression).
   private static Value evaluateRexp(Tree tree, Environment env) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      boolean result = false;

      String op = tree.getElement();

      Value valueL = evaluateExp(tree.getSubTree(0), env);
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(0) + "\n");
      }

      Value valueR = evaluateExp(tree.getSubTree(1), env);
      if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(1) + "\n");
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( op.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( op.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( op.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( op.equals(">=") )
      {
         result = resultL >= resultR;
      }
      else if ( op.equals("==") )
      {
         result = resultL == resultR;
      }
      else if ( op.equals("!=") )
      {
         result = resultL != resultR;
      }

      return new Value( result );
   }//evaluateRexp()


   // Evaluate an arithmetic expression.
   private static Value evaluateAexp(Tree tree, Environment env) throws EvalException
   {
      int result = 0;

      String node = tree.getElement();

      Value valueL = evaluateExp(tree.getSubTree(0), env);
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(0) + "\n");
      }
      int resultL = valueL.valueI;
      int resultR = 0;

      Value valueR = null;
      if ( tree.degree() >= 2 )
      {
         valueR = evaluateExp(tree.getSubTree(1), env);
         if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
         {
            throw new EvalException("not a integer expression: "
                                     + tree.getSubTree(1) + "\n");
         }
         resultR = valueR.valueI;
      }

      if ( node.equals("+") )
      {
         if ( tree.degree() == 1 )
            result = resultL;
         else
         {
            result = resultL + resultR;

            for (int i = 2; i < tree.degree(); i++)
            {
               Value temp = evaluateExp(tree.getSubTree(i), env);
               if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
               {
                  throw new EvalException("not a integer expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result += temp.valueI;
            }
         }
      }
      else if ( node.equals("-") )
      {
         if ( 2 < tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         if ( tree.degree() == 1 )
            result = -resultL;
         else
            result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         if ( 1 == tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         result = resultL * resultR;

         for (int i = 2; i < tree.degree(); i++)
         {
            Value temp = evaluateExp(tree.getSubTree(i), env);
            if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
            {
               throw new EvalException("not a integer expression: "
                                        + tree.getSubTree(i) + "\n");
            }
            result *= temp.valueI;
         }
      }
      else if ( node.equals("/") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = resultL % resultR;
      }
      else if ( node.equals("^") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = (int)Math.pow(resultL, resultR);
      }

      return new Value( result );
   }//evaluateAexp()
}
